package tree

import (
	"bytes"
	"errors"
	"math"
	"sort"

	"github.com/ethereum/go-ethereum/common"
	"github.com/ethereum/go-ethereum/crypto"
)

func combinedHash(first, second common.Hash) (hash common.Hash) {
	var buffer [64]byte
	if bytes.Compare(first[:], second[:]) <= 0 {
		copy(buffer[:32], first[:])
		copy(buffer[32:], second[:])
		copy(hash[:], crypto.Keccak256(buffer[:]))
	} else {
		copy(buffer[:32], second[:])
		copy(buffer[32:], first[:])
		copy(hash[:], crypto.Keccak256(buffer[:]))
	}
	return hash
}

type MerkleTree struct {
	layers                     [][]common.Hash
	proofs                     Proofs
	bufferElementPositionIndex map[common.Hash]int
	treeNode                   *MerkleTreeNode
}

type MerkleTreeNode struct {
	Hash  common.Hash
	Left  *MerkleTreeNode
	Right *MerkleTreeNode
}

func NewMerkleTree(proofs Proofs) (*MerkleTree, error) {
	sort.Sort(proofs)
	proofs = proofs.Distinct()

	bufferElementPositionIndex := make(map[common.Hash]int)
	for idx, el := range proofs {
		bufferElementPositionIndex[el] = idx
	}

	var err error
	tree := MerkleTree{
		proofs:                     proofs,
		bufferElementPositionIndex: bufferElementPositionIndex,
	}
	tree.layers, err = tree.GetLayers(proofs)
	if err != nil {
		return nil, err
	}
	tree.buildTree()

	return &tree, nil
}

func (m *MerkleTree) GetRoot() common.Hash {
	return m.layers[len(m.layers)-1][0]
}

func (m *MerkleTree) GetProof(el common.Hash) ([]common.Hash, error) {
	idx, ok := m.bufferElementPositionIndex[el]
	if !ok {
		return nil, errors.New("element does not exist in Merkle tree")
	}

	proof := make(Proofs, 0)
	for _, layer := range m.layers {
		pairElement, ok := m.getPairElement(idx, layer)
		if ok {
			proof = append(proof, pairElement)
		}
		idx = int(math.Floor(float64(idx) / 2))
	}

	return proof, nil
}

func (m *MerkleTree) GetLayers(proofs Proofs) ([][]common.Hash, error) {
	if len(proofs) == 0 {
		return nil, errors.New("empty tree")
	}

	layers := make([][]common.Hash, 0)
	layers = append(layers, proofs)

	for len(layers[len(layers)-1]) > 1 {
		layers = append(layers, m.GetNextLayer(layers[len(layers)-1]))
	}
	return layers, nil
}

func (m *MerkleTree) GetNextLayer(proofs Proofs) Proofs {
	layer := make(Proofs, 0)
	for idx, el := range proofs {
		if idx%2 == 0 {
			if idx+1 >= len(proofs) {
				layer = append(layer, el)
			} else {
				layer = append(layer, combinedHash(el, proofs[idx+1]))
			}
		}
	}
	return layer
}

func (m *MerkleTree) getPairElement(idx int, layer Proofs) (common.Hash, bool) {
	pairIdx := idx - 1
	if idx%2 == 0 {
		pairIdx = idx + 1
	}

	if pairIdx >= len(layer) {
		return common.Hash{}, false
	}

	return layer[pairIdx], true
}

func (m *MerkleTree) buildTree() {
	nodes := make([][]common.Hash, 0)
	for i := len(m.layers) - 1; i >= 0; i-- {
		if len(m.layers[i])%2 == 1 && i == 0 {
			nodes = append(nodes, append(m.layers[i], common.Hash{}))
		} else {
			nodes = append(nodes, m.layers[i])
		}
	}
	m.treeNode = buildTree(nodes)
}

func (m *MerkleTree) GetRootNode() *MerkleTreeNode {
	return m.treeNode
}

func (m *MerkleTree) FindMerkleNode(hash common.Hash) *MerkleTreeNode {
	if m.treeNode == nil {
		return nil
	}

	queue := []*MerkleTreeNode{m.treeNode}

	for len(queue) > 0 {
		size := len(queue)

		for i := 0; i < size; i++ {
			currentNode := queue[0]
			queue = queue[1:]

			if currentNode.Hash.Hex() == hash.Hex() {
				return currentNode
			}

			if currentNode.Left != nil {
				queue = append(queue, currentNode.Left)
			}
			if currentNode.Right != nil {
				queue = append(queue, currentNode.Right)
			}
		}
	}
	return nil
}

func MerkleTreeTraversal(root *MerkleTreeNode, depth int) (retNodes [][]common.Hash) {
	if root == nil {
		return nil
	}

	queue := []*MerkleTreeNode{root}

	currentDepth := 1

	for len(queue) > 0 {
		size := len(queue)

		layerNodes := make([]common.Hash, 0)
		for i := 0; i < size; i++ {
			currentNode := queue[0]
			queue = queue[1:]

			layerNodes = append(layerNodes, currentNode.Hash)

			if currentNode.Left != nil {
				queue = append(queue, currentNode.Left)
			}
			if currentNode.Right != nil {
				queue = append(queue, currentNode.Right)
			}
		}
		retNodes = append(retNodes, layerNodes)
		currentDepth++
		if currentDepth > depth {
			break
		}
	}
	return
}

func buildTree(nodes [][]common.Hash) *MerkleTreeNode {
	if len(nodes) == 0 {
		return nil
	}
	root := &MerkleTreeNode{Hash: nodes[0][0]}
	queue := []*MerkleTreeNode{root}

	for i := 1; i < len(nodes); i++ {
		var levelNodes []*MerkleTreeNode
		for j := 0; j < len(nodes[i]); j += 2 {
			current := queue[0]
			queue = queue[1:]

			current.Left = &MerkleTreeNode{Hash: nodes[i][j]}
			levelNodes = append(levelNodes, current.Left)

			if j+1 < len(nodes[i]) {
				current.Right = &MerkleTreeNode{Hash: nodes[i][j+1]}
				levelNodes = append(levelNodes, current.Right)
			}
		}
		queue = append(queue, levelNodes...)
	}
	return root
}
